# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookiescope cookie types and functions.
"""

from dataclasses import dataclass
from time import gmtime, strftime
from typing import Iterable
from urllib.parse import quote

from cookiescope.utility import abort, warning


@dataclass
class CookieData:
    """
    Cookie data class.

    All strings are unquoted (URL quoting escape sequences replaced).

    Field order below also determines display order.
    """
    #: Full or partial domain name.
    domain: str
    #: Cookie name.
    name: str
    #: Cookie path.
    path: str
    #: Cookie value (URL-unquoted data string).
    value: str
    #: HTTP-only flag.
    http_only: bool
    #: Is-secure flag.
    secure: bool
    #: Expiration timestamp (Unix UTC seconds since 1970).
    expires: int
    #: Creation timestamp (Unix UTC seconds since 1970 or 0).
    created: int

    def as_strings(self) -> Iterable[tuple[str, str]]:
        """
        Convert fields to (name, value) string pairs

        Returns:
            iterable (name, value) string pairs
        """
        yield 'domain', self.domain
        yield 'name', self.name
        yield 'path', self.path
        yield 'value', self.value
        yield 'http_only', str(self.http_only).lower()
        yield 'secure', str(self.secure).lower()
        if self.expires:
            yield 'expires', strftime('%c', gmtime(self.expires))
        if self.created:
            yield 'created', strftime('%c', gmtime(self.created))

    def as_cookie_file_line(self):
        """
        Convert to Netscape cookie file format line.

        See https://curl.se/docs/http-cookies.html.

        Returns:
            cookie file line
        """
        return '\t'.join([
            f'#HttpOnly_{self.domain}' if self.http_only else self.domain,
            'TRUE',     # subdomains
            self.path,
            'TRUE' if self.secure else 'FALSE',
            str(self.expires),
            self.name,
            quote(self.value),
        ])


# --- Types.
#: Filter field name and possible values handled with logical OR.
Filter = tuple[str, list[str]]
#: Multiple filters handled with logical AND.
FilterBy = Iterable[Filter]
#: Sorting specified as a sequence of field names.
SortBy = Iterable[str]

# --- Constants.
#: Field names supported for filtering.
FILTER_FIELDS = ['domain', 'name', 'path', 'value']
#: Field names supported for sorting.
SORT_FIELDS = ['domain', 'name', 'path', 'value']
#: Default sort fields.
DEFAULT_SORT_FIELDS = ['domain', 'path']


def get_filter_by(filter_exprs: Iterable[str]) -> FilterBy:
    """
    Convert name=value raw filter expressions to filter-by list.

    Values strings are HTTP-escaped. Multiple values may be comma-separated.

    Args:
        filter_exprs: name=value filter expressions

    Returns:
        dictionary mapping attribute names to lists of partial values
    """
    filters: list[Filter] = []
    for filter_expr in filter_exprs:
        expr_parts = filter_expr.split('=', maxsplit=1)
        if len(expr_parts) != 2 or not expr_parts[1] or expr_parts[0] not in FILTER_FIELDS:
            abort(f'Bad name=value filter expression: {filter_expr}')
        name, value = expr_parts
        filters.append((name, value.split(',')))
    return filters


def filter_cookies(unfiltered_cookies: Iterable[CookieData],
                   filter_by: FilterBy | None,
                   ) -> Iterable[CookieData]:
    """
    Filter cookies.

    Filters are ANDed together. Filter values are ORed together.

    Args:
        unfiltered_cookies: unfiltered input cookies
        filter_by: optional (name, values) filter pairs

    Returns:
        iterable filtered cookies
    """
    if filter_by is None:
        return unfiltered_cookies
    # Lower-case the filter values for normalized comparison.
    filter_by = [(name, [value.lower() for value in values]) for name, values in filter_by]
    for cookie in unfiltered_cookies:
        for name, values in filter_by:
            if name not in FILTER_FIELDS:
                warning(f'Ignoring bad filter field name: {name}')
                break
            for value in values:
                # Normalize comparison by lower-casing cookie values.
                if getattr(cookie, name).lower().find(value) == -1:
                    break
            else:
                continue
            break
        else:
            yield cookie


def sort_cookies(unsorted_cookies: Iterable[CookieData],
                 sort_by: SortBy | None,
                 ) -> Iterable[CookieData]:
    """
    Sort cookies.

    Args:
        unsorted_cookies: unsorted input cookies
        sort_by: optional attribute names to sort by in priority order

    Returns:
        iterable sorted cookies
    """
    if not sort_by:
        return unsorted_cookies
    bad_fields: list[str] = []
    sort_fields: list[str] = []
    for sort_field in sort_by:
        if sort_field in SORT_FIELDS and sort_field not in sort_fields:
            sort_fields.append(sort_field)
        else:
            bad_fields.append(sort_field)
    if bad_fields:
        warning(f'Ignoring bad sort field(s): {" ".join(bad_fields)}')
    if not sort_fields:
        return unsorted_cookies

    def get_sort_values(cookie: CookieData) -> list[str]:
        return [getattr(cookie, name) for name in sort_fields]

    return sorted(unsorted_cookies, key=get_sort_values)


def display_cookies(cookies: Iterable[CookieData], heading: str = None):
    """
    Display cookies.

    Args:
        cookies: iterable cookies to display
        heading: optional heading to display
    """
    if heading:
        print(f'=== {heading} ===')
    for cookie in cookies:
        print('')
        for name, value in cookie.as_strings():
            print(f'{name}={value}')


def display_cookie_jar(cookies: Iterable[CookieData]):
    """
    Display cookie jar, i.e. Netscape format text cookies.

    https://curl.se/docs/http-cookies.html

    Args:
        cookies: cookies to display
    """

    for cookie in cookies:
        print(cookie.as_cookie_file_line())



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

import sqlite3
from abc import ABC, abstractmethod
from collections import namedtuple
from pathlib import Path
from typing import Iterable, Iterator
from urllib.parse import unquote

from cookiescope.cookies import CookieData, FilterBy, SortBy, filter_cookies, sort_cookies
from cookiescope.utility import open_binary_file


class SQLiteCookiesBase(ABC):
    """Base utility class for SQLite cookie access."""

    #: Named tuple that row data is assigned to.
    CanonicalRow = namedtuple(
        'CanonicalRow',
        [
            'domain',
            'name',
            'value',
            'path',
            'has_expires',
            'expires_utc',
            'creation_utc',
            'httponly',
            'secure',
        ]
    )

    #: Must be provided by the subclass.
    DatabaseRow: namedtuple = None

    def __init__(self, path: Path, table_name: str):
        """
        SQLiteCookiesBase constructor.

        Args:
            path: database file path
            table_name: cookies table name
        """
        self.path = path
        self.table_name = table_name
        assert self.DatabaseRow is not None
        # noinspection PyProtectedMember
        columns_string = ", ".join(self.DatabaseRow._fields)
        self.cookie_query = f'SELECT {columns_string} FROM {self.table_name}'

    def is_cookies_db(self) -> bool:
        """
        Check if file is a SQLite cookies database.

        Returns:
            True if file is a SQLite database
        """
        # First sanity-check that the file can be opened at all.
        with open_binary_file(self.path):
            pass
        # Then see if we can query it as a cookies database.
        try:
            connection = sqlite3.connect(self.path)
            try:
                cursor = connection.execute(
                    f"SELECT count(*) from sqlite_master "
                    f"WHERE type = 'table' and name = '{self.table_name}'"
                )
                try:
                    return cursor.fetchone()[0] == 1
                finally:
                    cursor.close()
            finally:
                connection.close()
        except sqlite3.Error:
            return False

    @abstractmethod
    def canonicalize_row(self, row: DatabaseRow) -> CanonicalRow:
        """
        Required method to convert database row to canonical row.
        Args:
            row: database row

        Returns:
            canonical row
        """
        ...

    def generate_cookies(self,
                         filter_by: FilterBy,
                         sort_by: SortBy,
                         ) -> Iterable[CookieData]:
        """Required override: query cookies with optional filtering and sorting."""
        def _generate() -> Iterator[CookieData]:
            connection = sqlite3.connect(self.path)
            try:
                cursor = connection.cursor()
                try:
                    cursor.execute(self.cookie_query)
                    for raw_row in cursor:
                        row = self.canonicalize_row(self.DatabaseRow(*raw_row))
                        yield CookieData(
                            domain=row.domain,
                            name=row.name,
                            path=row.path,
                            value=unquote(row.value),
                            http_only=bool(row.httponly),
                            secure=bool(row.secure),
                            expires=row.expires_utc if row.has_expires else 0,
                            created=row.creation_utc,
                        )
                finally:
                    cursor.close()
            finally:
                connection.close()
        # Avoid excess memory consumption by passing generator to filtering/sorting.
        filtered_cookies = filter_cookies(_generate(), filter_by)
        return sort_cookies(filtered_cookies, sort_by)



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

# Decryption is based on code from the following project. But the code has been
# heavily-modified and reorganized.
#
# https://github.com/n8henrie/pycookiecheat
#
# That project's code is covered by the following license:
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Nathan Henrie
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import keyring

"""
Cookiescope Linux decryptor.
"""

from cookiescope.utility import abort
from .posix import PosixDecryptorBase


class LinuxDecryptor(PosixDecryptorBase):
    """Linux decryptor."""

    #: Decryption iteration count.
    iterations = 1

    def get_password(self) -> str:
        """Get storage password - see base class method docstring."""
        password = gnome_get_password(self.browser_name)
        # Try to get password from keyring, which should support KDE / KWallet
        # if dbus-python is installed.
        if password is None:
            password = kde_get_password(self.browser_name)
        if password is None:
            abort('Unable to get storage password.')
        return password


def gnome_get_password(browser_name: str) -> str | None:
    """
    Gnome get storage password.

    Try to get pass from Gnome / libsecret if it seems available
    https://github.com/n8henrie/pycookiecheat/issues/12

    Args:
        browser_name: browser name

    Returns:
        password or None if not available
    """
    keyring_label = f'{browser_name} Safe Storage'
    keyring_item_name = browser_name
    keyring_app_name = browser_name.lower()
    try:
        import gi
        gi.require_version("Secret", "1")
        from gi.repository import Secret
    except ImportError:
        return None
    keyring_attr_name = keyring_item_name
    flags = Secret.ServiceFlags.LOAD_COLLECTIONS
    service = Secret.Service.get_sync(flags)
    gnome_keyring = service.get_collections()
    unlocked_keyrings = service.unlock_sync(gnome_keyring).unlocked
    for unlocked_keyring in unlocked_keyrings:
        for item in unlocked_keyring.get_items():
            if item.get_label() == keyring_label:
                item_app = item.get_attributes().get('application', keyring_attr_name)
                if item_app.lower() == keyring_app_name:
                    item.load_secret_sync()
                    return item.get_secret().get_text()


def kde_get_password(browser_name: str) -> str | None:
    """
    KDE/KWallet get storage password.

    Args:
        browser_name: browser name

    Returns:
        password or None if not available
    """
    keyring_label = f'{browser_name} Safe Storage'
    keyring_item_name = browser_name
    # Should support KDE / KWallet if dbus-python is installed.
    try:
        service_name = f'{keyring_item_name} Keys'
        return keyring.get_password(service_name, keyring_label)
    except RuntimeError:
        return None



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

# Decryption is based on code from the following project. But the code has been
# heavily-modified and reorganized.
#
# https://github.com/n8henrie/pycookiecheat
#
# That project's code is covered by the following license:
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Nathan Henrie
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Cookiescope MacOS decryptor.
"""

import keyring

from .posix import PosixDecryptorBase


class MacOSDecryptor(PosixDecryptorBase):
    """MacOS decryptor."""

    #: Decryption iteration count.
    iterations = 1003

    def get_password(self) -> str:
        """Get storage password - see base class method docstring."""
        return keyring.get_password(f'{self.browser_name} Safe Storage', self.browser_name)



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookiescope base decryptor.
"""

from abc import ABC, abstractmethod


class DecryptorBase(ABC):
    """Abstract decryptor class."""

    def __init__(self, browser_name: str):
        """
        Base decryptor constructor.

        Args:
            browser_name: browser name
        """
        self.browser_name = browser_name

    @abstractmethod
    def decrypt(self, encrypted_value: bytes) -> str:
        """
        Decrypt a value.

        Args:
            encrypted_value: encrypted value

        Returns:
            decrypted value
        """
        ...



#!/usr/bin/env python3
#
# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookie scope.

Queries and displays browser cookies.

See command line help text for more information.

Binary cookies parsing code was originally based on this project:

https://github.com/ktnjared/BinaryCookieReader.git

It has been heavily modified and expanded here. The original code is probably
not recognizable at this point.
"""

import argparse
import os
from pathlib import Path

from cookiescope.browsers import (
    BrowserBase,
    ChromeBrowser,
    ChromiumBrowser,
    EdgeBrowser,
    FirefoxBrowser,
    GenericChromeBrowser,
    SafariBrowser,
)
from cookiescope.cookies import (
    DEFAULT_SORT_FIELDS,
    display_cookies,
    display_cookie_jar,
    get_filter_by,
)
from cookiescope.utility import abort


#: Command line help description.
CLI_DESCRIPTION = 'Cookie query tool.'
#: Command line help epilog text.
CLI_EPILOG = '''
Cookie source can be a browser name, e.g. "safari", or a cookies file path. A
browser name can followed by ":<profile>" to specify a user profile. E.g.
"firefox:lucy" targets the Firefox profile named "lucy". Otherwise it works with
the default profile.

Filter values match full or partial attribute values. Multiple filter values can
be comma-separated. For the comparison, filter values are HTTP-quoted and
lower-cased, and cookie fields are HTTP-unquoted and also lower-cased.

If you have problems accessing an existing cookies file try quitting all running
browser processes. Some processes may be running in the background, even when
there are no visible windows.

Frequently-used attribute names:
  * value - primary cookie data field
  * name - cookie name
  * domain - website domain responsible for cookie
  * path - website location path string

The following named browsers are supported:
'''.strip()
#: Cookie source argument help.
COOKIE_SOURCE_HELP = 'cookies path or browser[:profile]'
FILTER_HELP = 'name=value expression for filtering on cookie fields'


NAMED_BROWSERS: dict[str, type[BrowserBase]] = {
    'chrome': ChromeBrowser,
    'chromium': ChromiumBrowser,
    'edge': EdgeBrowser,
    'firefox': FirefoxBrowser,
    'safari': SafariBrowser,
}

FILE_CHECK_BROWSERS: list[type[BrowserBase]] = [
    GenericChromeBrowser,
    FirefoxBrowser,
    SafariBrowser,
    ChromiumBrowser,
]


def get_browser_for_cookie_source(cookie_source: str) -> BrowserBase:
    """
    Get browser object based on cookie source (file path or browser name).

    Args:
        cookie_source: file path or browser name

    Returns:
        browser object for processing query
    """
    if os.path.isfile(cookie_source):
        file_path = Path(cookie_source)
        for browser_class in FILE_CHECK_BROWSERS:
            browser = browser_class.from_file(file_path)
            if browser is not None:
                return browser
        abort('Unable to identify browser for file provided.')
    if os.path.sep in cookie_source:
        abort('File not found.')
    cookie_source_parts = cookie_source.lower().split(':', maxsplit=1)
    if len(cookie_source_parts) == 2:
        browser_name, profile = cookie_source_parts
    else:
        browser_name, profile = cookie_source_parts[0], None
    browser_name = browser_name.lower()
    if browser_name not in NAMED_BROWSERS:
        abort(f'Browser not supported: {browser_name}')
    browser_class = NAMED_BROWSERS[browser_name]
    file_path = browser_class.find_cookies(profile)
    if file_path is None:
        abort(f'{browser_class.name} cookies file not found.')
    browser = browser_class.from_file(file_path)
    if browser is None:
        abort(f'{browser_class.name} browser class did not recognize cookies file.',
              str(file_path))
    return browser


def main():
    """Main function."""
    epilog_parts = [CLI_EPILOG] + [f'  - {name}' for name in sorted(NAMED_BROWSERS.keys())]
    arg_parser = argparse.ArgumentParser(
        description=CLI_DESCRIPTION,
        epilog=os.linesep.join(epilog_parts),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    arg_parser.add_argument(dest='COOKIE_SOURCE', help=COOKIE_SOURCE_HELP)
    arg_parser.add_argument(dest='FILTER', nargs='*', help=FILTER_HELP)
    arg_parser.add_argument('-j', '--jar', dest='JAR', action='store_true',
                            help='generate Netscape cookie jar format, e.g. for use with "curl"')
    args = arg_parser.parse_args()
    filter_by = get_filter_by(args.FILTER)
    browser = get_browser_for_cookie_source(args.COOKIE_SOURCE)
    cookies = browser.generate_cookies(filter_by=filter_by, sort_by=DEFAULT_SORT_FIELDS)
    if args.JAR:
        display_cookie_jar(cookies)
    else:
        display_cookies(cookies, heading=f'{browser.name}: {browser.file_path}')


if __name__ == '__main__':
    main()



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookiescope generic Chrome browser support.
"""

# Decryption is based on code from the following project. But the code has been
# heavily-modified and reorganized.
#
# https://github.com/n8henrie/pycookiecheat
#
# That project's code is covered by the following license:
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Nathan Henrie
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
from collections import namedtuple
from pathlib import Path
from typing import Iterable, Self

from cookiescope.cookies import CookieData, FilterBy, SortBy
from cookiescope.extractors import SQLiteCookiesBase
from .base import BrowserBase, LocationMap


class GenericChromeSQLiteCookies(SQLiteCookiesBase):

    #: Required database row named tuple.
    DatabaseRow = namedtuple(
        'DatabaseRow',
        [
            'host_key',
            'name',
            'value',
            'encrypted_value',
            'path',
            'has_expires',
            'expires_utc',
            'creation_utc',
            'is_httponly',
            'is_secure',
        ],
    )

    # Ignore unresolved references due to excluded platform-specific code.
    # noinspection PyUnresolvedReferences
    def __init__(self, path: Path, name: str):
        """
        SQLiteCookies constructor.

        Args:
            path: cookies database file path
            name: browser name
        """
        if sys.platform == 'darwin':
            from cookiescope.decryptors.macos import MacOSDecryptor
            self.decryptor = MacOSDecryptor(name)
        elif sys.platform == 'linux':
            from cookiescope.decryptors.linux import LinuxDecryptor
            self.decryptor = LinuxDecryptor(name)
        elif sys.platform == 'win32':
            from cookiescope.decryptors.windows import WindowsDecryptor
            # Windows decryption needs to open the state file, which is 3 directories
            # above the cookies database for Chrome-based browsers.
            state_file = path.parent.parent.parent / 'Local State'
            self.decryptor = WindowsDecryptor(name, state_file)
        else:
            raise ValueError(f'Unsupported platform for {name} browser: {sys.platform}')
        super().__init__(path, 'cookies')

    def canonicalize_row(self, row: DatabaseRow) -> SQLiteCookiesBase.CanonicalRow:
        if row.encrypted_value:
            value = self.decryptor.decrypt(row.encrypted_value)
        else:
            value = row.value
        return SQLiteCookiesBase.CanonicalRow(
            domain=row.host_key,
            name=row.name,
            value=value,
            path=row.path,
            has_expires=row.has_expires,
            expires_utc=row.expires_utc / 1000000 - 11644473600,
            creation_utc=row.creation_utc / 1000000 - 11644473600,
            httponly=row.is_httponly,
            secure=row.is_secure,
        )


class GenericChromeBrowser(BrowserBase):
    """Generic Chrome-based browser implementation."""

    name = 'Chrome (generic)'

    # Should be overridden.
    db_paths: LocationMap = {}

    def __init__(self, file_path: Path, cookies_db: GenericChromeSQLiteCookies):
        """
        Generic Chrome browser base constructor.

        Args:
            file_path: cookies file or database path
        """
        assert self.db_paths
        super().__init__(file_path)
        self.cookies_db = cookies_db

    @classmethod
    def from_file(cls, path: Path) -> Self | None:
        """Required conditional factory method."""
        cookies_db = GenericChromeSQLiteCookies(path, cls.name)
        if not cookies_db.is_cookies_db():
            return None
        return cls(path, cookies_db)

    @classmethod
    def find_cookies(cls, profile: str | None) -> Path | None:
        """Required override to locate the cookies database."""
        return cls.find_file(cls.db_paths)

    def generate_cookies(self, filter_by: FilterBy, sort_by: SortBy) -> Iterable[CookieData]:
        """Required override to generate cookies."""
        return self.cookies_db.generate_cookies(filter_by, sort_by)



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

# Decryption is based on code from the following project. But the code has been
# heavily-modified and reorganized.
#
# https://github.com/n8henrie/pycookiecheat
#
# That project's code is covered by the following license:
#
# The MIT License (MIT)
#
# Copyright (c) 2015 Nathan Henrie
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Cookiescope common Posix, i.e. Linux and MacOS, decryption support.
"""

from abc import ABC, abstractmethod
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.hashes import SHA1
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from .base import DecryptorBase


class PosixDecryptorBase(DecryptorBase, ABC):
    """Posix decryptor base class."""

    #: Decryption iteration count (subclass must provide).
    iterations: int = None

    def __init__(self, browser_name: str):
        """
        Linux/MacOS decryptor base class constructor.

        Args:
            browser_name: browser name
        """
        assert self.iterations
        super().__init__(browser_name)
        kdf = PBKDF2HMAC(algorithm=SHA1(), iterations=self.iterations, length=16, salt=b'saltysalt')
        password = self.get_password()
        encryption_key = kdf.derive(password.encode('utf8'))
        algorithm = AES(encryption_key)
        self.cipher = Cipher(algorithm=algorithm, mode=CBC(b' ' * 16))

    @abstractmethod
    def get_password(self) -> str:
        """
        Required method to get the storage password.

        Returns:
            password
        """
        ...

    def decrypt(self, encrypted_value: bytes) -> str:
        """
        Common Linux/MacOS method to decrypt a value.

        Args:
            encrypted_value: encrypted value

        Returns:
            decrypted value
        """
        # Encrypted cookies should be prefixed with 'v10' or 'v11' according to the
        # Chromium code. Strip it off.
        encrypted_value = encrypted_value[3:]
        decryptor = self.cipher.decryptor()
        decrypted = decryptor.update(encrypted_value) + decryptor.finalize()
        # Strip padding and decode.
        last = decrypted[-1]
        if isinstance(last, int):
            return decrypted[:-last].decode('utf8')
        # noinspection PyTypeChecker
        return decrypted[: -ord(last)].decode('utf8')



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookie scope utilities.
"""

import sys
from contextlib import contextmanager
from pathlib import Path
from typing import IO, Iterator


def warning(*messages: str):
    """
    Display warning messages.

    Args:
        messages: warning messages
    """
    for message in messages:
        sys.stderr.write(f'WARNING: {message}\n')


def error(*messages: str):
    """
    Display error messages.

    Args:
        messages: error messages
    """
    for message in messages:
        sys.stderr.write(f'ERROR: {message}\n')


def abort(*messages: str):
    """
    Display error messages and quit.

    Args:
        messages: error messages
    """
    error(*messages)
    sys.exit(1)


@contextmanager
def open_binary_file(path: Path) -> Iterator[IO]:
    try:
        with path.open('rb') as file:
            yield file
    except PermissionError:
        abort('Permission denied opening file:',
              str(path),
              'Make sure there are no foreground or background browser processes.')
    except (IOError, OSError) as exc:
        abort('Unable to open file due to exception:', str(path), str(exc))



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Binary cookies handling.
"""

from io import BytesIO
from pathlib import Path
from struct import unpack
from urllib.parse import unquote

from cookiescope.cookies import CookieData
from cookiescope.utility import abort, open_binary_file
from typing import AnyStr, IO, Iterator, Self


class BinaryCookiesExtractor:
    """Binary cookies file reader and data extractor/decoder."""

    def __init__(self, stream: IO):
        """
        Binary cookies extractor constructor.

        Args:
            stream: file stream to read cookies from
        """
        self._stream = stream

    def get_bytes(self, length: int, expect: bytes = None) -> AnyStr:
        """
        Extract bytes from stream.

        Args:
            length: number of bytes
            expect: optional expected value to check

        Returns:
            extracted bytes
        """
        value = self._stream.read(length)
        if expect is not None and value != expect:
            abort(f'Expected bytes "{expect}", found "{value}".')
        return value

    def get_string(self, length: int, encoding: str = 'utf-8', expect: str = None) -> str:
        """
        Extract string from stream.

        Args:
            length: number of characters
            encoding: optional encoding (default: utf-8)
            expect: optional expected value to check

        Returns:
            extracted string value
        """
        raw_value = self.get_bytes(length)
        value = raw_value.decode(encoding)
        if expect is not None and value != expect:
            abort(f'Expected string "{expect}", found "{value}".')
        return value

    def get_header_integer(self) -> int:
        """
        Extract integer from stream header (big-endian integer format).

        Returns:
            extracted integer value
        """
        return unpack('>i', self.get_bytes(4))[0]

    def get_integer(self) -> int:
        """
        Extract integer from stream body (little-endian integer format).

        Returns:
            extracted integer value
        """
        return unpack('<i', self.get_bytes(4))[0]

    def get_flags(self) -> tuple[bool, bool]:
        """
        Extract flag string from stream by checking flag integer bits.

        Returns:
            extracted (secure, http_only) boolean flags
        """
        flags = self.get_integer()
        return bool(flags & 0x00000001), bool(flags & 0x00000004)

    def get_attribute(self) -> str:
        """
        Extract attribute value from stream.

        Returns:
            extracted attribute value
        """
        value = ''
        byte = self.get_bytes(1)
        while unpack('<b', byte)[0] != 0:
            value += byte.decode('utf-8')
            byte = self.get_bytes(1)
        return value

    def get_date_time(self) -> int:
        """
        Extract date/time string value from stream.

        Returns:
            extracted seconds since 1970 (Unix epoch) as integer
        """
        # Expiry date is in Mac epoch format: Starts from 1/Jan/2001 (978307200 in seconds)
        return int(unpack('<d', self.get_bytes(8))[0] + 978307200)

    def get_block(self, length) -> Self:
        """
        Get sub-extractor for stream block.

        Args:
            length: block length

        Returns:
            block sub-extractor
        """
        return self.__class__(BytesIO(self.get_bytes(length)))

    def set_offset(self, offset: int):
        """
        Set stream offset (position).

        Args:
            offset: byte offset
        """
        self._stream.seek(offset)


def is_binary_cookies_file(path: Path) -> bool:
    """
    Check if file appears to be a binary cookies file.

    Args:
        path: file path

    Returns:
        True if the file contains binary cookies
    """
    with open_binary_file(path) as binary_file:
        extractor = BinaryCookiesExtractor(binary_file)
        if extractor.get_string(4, encoding='ascii') != 'cook':
            return False
        return True


def generate_binary_cookies(path: Path) -> Iterator[CookieData]:
    """
    Generate cookies by reading file and extracting individual cookies.

    Args:
        path: file path

    Returns:
        cookie iterator
    """
    with open_binary_file(path) as binary_file:
        extractor = BinaryCookiesExtractor(binary_file)
        extractor.get_string(4, encoding='ascii', expect='cook')
        num_pages = extractor.get_header_integer()
        page_sizes = [extractor.get_header_integer() for _idx in range(num_pages)]
        page_blocks = [extractor.get_block(length) for length in page_sizes]
        for page_block in page_blocks:
            page_block.get_bytes(4, expect=bytes([0, 0, 1, 0]))
            num_cookies = page_block.get_integer()
            cookie_offsets = [page_block.get_integer() for _idx in range(num_cookies)]
            page_block.get_bytes(4, expect=bytes([0, 0, 0, 0]))
            for offset in cookie_offsets:
                page_block.set_offset(offset)
                cookie_size = page_block.get_integer()
                cookie_block = page_block.get_block(cookie_size)
                cookie_block.get_bytes(4)  # unknown
                secure, http_only = cookie_block.get_flags()
                cookie_block.get_bytes(4)  # unknown
                url_offset = cookie_block.get_integer()
                name_offset = cookie_block.get_integer()
                path_offset = cookie_block.get_integer()
                value_offset = cookie_block.get_integer()
                cookie_block.get_bytes(8)  # end of cookie
                expires = cookie_block.get_date_time()
                created = cookie_block.get_date_time()
                cookie_block.set_offset(url_offset - 4)
                domain = cookie_block.get_attribute()
                cookie_block.set_offset(name_offset - 4)
                name = cookie_block.get_attribute()
                cookie_block.set_offset(path_offset - 4)
                path = cookie_block.get_attribute()
                cookie_block.set_offset(value_offset - 4)
                value = unquote(cookie_block.get_attribute())
                yield CookieData(
                    domain=domain,
                    name=name,
                    path=path,
                    value=value,
                    http_only=http_only,
                    secure=secure,
                    expires=expires,
                    created=created,
                )



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookiescope Microsoft Edge browser support.
"""

from .base import LocationMap
from .chrome_generic import GenericChromeBrowser


class EdgeBrowser(GenericChromeBrowser):
    """Microsoft Edge browser implementation."""
    name = 'Microsoft Edge'
    db_paths: LocationMap = {
        'darwin': [
            '~/Library/Application Support/Microsoft Edge/Default/Cookies',
        ],
        'linux': [
            '~/.config/edge/Default/Cookies',
        ],
        'win32': [
            '~/AppData/Local/Microsoft/Edge/User Data/Default/Network/Cookies',
        ],
    }



#!/usr/bin/env python3
#
# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookiescope Safari browser support.
"""

from pathlib import Path
from typing import Iterable, Self

from cookiescope.extractors import generate_binary_cookies, is_binary_cookies_file
from cookiescope.cookies import CookieData, FilterBy, SortBy, filter_cookies, sort_cookies
from .base import BrowserBase, LocationMap


class SafariBrowser(BrowserBase):
    """Safari browser implementation."""

    name = 'Safari'

    cookies_paths: LocationMap = {
        'darwin': [
            '~/Library/Containers/com.apple.Safari/Data/Library/Cookies/Cookies.binarycookies',
            '~/Library/Cookies/Cookies.binarycookies',
        ],
    }

    @classmethod
    def from_file(cls, path: Path) -> Self | None:
        """Required conditional factory method."""
        return cls(path) if is_binary_cookies_file(path) else None

    @classmethod
    def find_cookies(cls, profile: str | None) -> Path | None:
        """Required method to locate the cookies file."""
        return cls.find_file(cls.cookies_paths)

    def generate_cookies(self, filter_by: FilterBy, sort_by: SortBy) -> Iterable[CookieData]:
        """Required override: query cookies with optional filtering and sorting."""
        filtered_cookies = filter_cookies(generate_binary_cookies(self.file_path), filter_by)
        return sort_cookies(filtered_cookies, sort_by)



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookiescope MacOS decryptor.
"""

# See https://stackoverflow.com/questions/60416350/chrome-80-how-to-decode-cookies

import base64
import json
import win32crypt
from Crypto.Cipher import AES
from pathlib import Path

from .base import DecryptorBase


class WindowsDecryptor(DecryptorBase):
    """Windows decryptor."""

    def __init__(self, browser_name: str, state_file: Path):
        """
        Base decryptor constructor.

        Args:
            browser_name: browser name
            state_file: state file path used for decryption
        """
        super().__init__(browser_name)
        self.state_file = state_file

    def decrypt(self, encrypted_value: bytes) -> str:
        """
        Windows method to decrypt a value.

        Args:
            encrypted_value: encrypted value

        Returns:
            decrypted value
        """
        with open(self.state_file, 'r') as file:
            encrypted_key = json.loads(file.read())['os_crypt']['encrypted_key']
        encrypted_key = base64.b64decode(encrypted_key)                                       # Base64 decoding
        encrypted_key = encrypted_key[5:]                                                     # Remove DPAPI
        decrypted_key = win32crypt.CryptUnprotectData(encrypted_key, None, None, None, 0)[1]  # Decrypt key
        # data = bytes.fromhex('763130...') # the encrypted cookie
        nonce = encrypted_value[3:3+12]
        ciphertext = encrypted_value[3+12:-16]
        tag = encrypted_value[-16:]
        cipher = AES.new(decrypted_key, AES.MODE_GCM, nonce=nonce)
        # noinspection PyTypeChecker
        return cipher.decrypt_and_verify(ciphertext, tag)



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookiescope Firefox browser support.
"""

from collections import namedtuple
from configparser import ConfigParser
from pathlib import Path
from typing import Iterable, Self

from cookiescope.cookies import CookieData, FilterBy, SortBy
from cookiescope.extractors import SQLiteCookiesBase
from cookiescope.utility import error
from .base import BrowserBase, LocationMap


class FirefoxBrowser(BrowserBase):
    """Firefox browser implementation."""

    name = 'Firefox'

    profiles_ini_paths: LocationMap = {
        'darwin': [
            '~/Library/Application Support/Firefox/profiles.ini',
        ],
        'linux': [
            '~/.mozilla/firefox/profiles.ini',
        ],
        'win32': [
            '~/AppData/Roaming/Mozilla/Firefox/profiles.ini',
        ],
    }

    class SQLiteCookies(SQLiteCookiesBase):

        DatabaseRow = namedtuple(
            'DatabaseRow',
            [
                'host',
                'name',
                'value',
                'path',
                'expiry',
                'creationTime',
                'isHttpOnly',
                'isSecure',
            ],
        )

        def __init__(self, path: Path):
            """
            SQLiteCookies constructor.

            Args:
                path: cookies database file path
            """
            super().__init__(path, 'moz_cookies')

        def canonicalize_row(self, row: DatabaseRow) -> SQLiteCookiesBase.CanonicalRow:
            return SQLiteCookiesBase.CanonicalRow(
                domain=row.host,
                name=row.name,
                value=row.value,
                path=row.path,
                has_expires=row.expiry,
                expires_utc=row.expiry,
                creation_utc=row.creationTime / 1000000,
                httponly=row.isHttpOnly,
                secure=row.isSecure,
            )

    def __init__(self, file_path: Path, cookies_db: SQLiteCookies):
        """
        Browser base constructor.

        Args:
            file_path: cookies file or database path
        """
        super().__init__(file_path)
        self.cookies_db = cookies_db

    @classmethod
    def from_file(cls, path: Path) -> Self | None:
        """Required conditional factory method."""
        cookies_db = cls.SQLiteCookies(path)
        if not cookies_db.is_cookies_db():
            return None
        return cls(path, cookies_db)

    @classmethod
    def find_cookies(cls, profile: str | None) -> Path | None:
        """Required override to find the cookies database file."""
        # NB: Firefox seems to be a bit flakey with how it handles finding the
        # active profile. So in some cases users may need to explicitly specify
        # a profile, because the default may not actually be active.
        profiles = ConfigParser()
        profiles_ini_path = cls.find_file(cls.profiles_ini_paths)
        if profiles_ini_path is None:
            error(f'Unable to find Firefox profiles configuration.')
            return None
        parsed_files = profiles.read(profiles_ini_path)
        if not parsed_files:
            error(f'Unable to parse Firefox profiles configuration: {profiles_ini_path}')
            return None
        # Find the cookies database path.
        if profile:
            # Match specific profile.
            for section in profiles.sections():
                if profiles.get(section, 'Name', fallback=None) == profile:
                    cookies_db_path = profiles.get(section, 'Path', fallback=None)
                    if cookies_db_path is not None:
                        if profiles.get(section, 'IsRelative', fallback=None):
                            return profiles_ini_path.parent / cookies_db_path
                        return Path(cookies_db_path)
                    return None
        # Otherwise look for default profile with cookie file.
        for section in profiles.sections():
            default = profiles.get(section, 'Default', fallback=None)
            if default:
                # Handle the complication of an "Install..." profile providing
                # the active path via the "Default" attribute. That path may be
                # relative without an "IsRelative" attribute to flag as such.
                if default != '1':
                    cookies_db_path = Path(default) / 'cookies.sqlite'
                    if cookies_db_path.is_file():
                        return cookies_db_path
                    return profiles_ini_path.parent / cookies_db_path
                profile_folder = profiles.get(section, 'Path', fallback=None)
                if profile_folder:
                    if profiles.get(section, 'IsRelative', fallback=None):
                        profile_folder = profiles_ini_path.parent / profile_folder
                    else:
                        profile_folder = Path(profile_folder)
                    cookies_db_path = profile_folder / 'cookies.sqlite'
                    if cookies_db_path.is_file():
                        return cookies_db_path
        return None

    def generate_cookies(self,
                         filter_by: FilterBy,
                         sort_by: SortBy,
                         ) -> Iterable[CookieData]:
        """Required override to generate cookies."""
        return self.cookies_db.generate_cookies(filter_by, sort_by)



# Copyright (C) 2023, Steven Cooper
#
# This file is part of Cookiescope.
#
# Cookiescope is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Cookiescope is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Cookiescope.  If not, see <https://www.gnu.org/licenses/>.

"""
Cookiescope base browser class.
"""

import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable, Self

from cookiescope.cookies import CookieData, FilterBy, SortBy

#: Mapping of platform to possible file locations.
LocationMap = dict[str, Iterable[str | Path]]


class BrowserBase(ABC):
    """Abstract browser class."""

    name: str = None

    def __init__(self, file_path: Path):
        """
        Browser base constructor.

        Args:
            file_path: cookies file or database path
        """
        self.file_path = file_path

    @classmethod
    @abstractmethod
    def from_file(cls, path: Path) -> Self | None:
        """
        Required conditional factory method..

        Args:
            path: file path

        Returns:
            browser instance if the file belongs to browser
        """
        ...

    @classmethod
    @abstractmethod
    def find_cookies(cls, profile: str | None) -> Path | None:
        """
        Required method to find the cookies file.

        Args:
            profile: optional profile name

        Returns:
            cookies file path if found
        """
        ...

    @abstractmethod
    def generate_cookies(self, filter_by: FilterBy, sort_by: SortBy) -> Iterable[CookieData]:
        """
        Required method to generate cookies with optional filtering and sorting.

        Args:
            filter_by: filters as a mapping of attribute names to filtered values
            sort_by: sort by named attributes in order provided
        """
        ...

    @classmethod
    def find_file(cls, location_map: LocationMap) -> Path | None:
        """
        Utility method to search for a file given multiple possible platform-specific paths.

        Args:
            location_map: possible file locations mapped by platform

        Returns:
            found path
        """
        if sys.platform not in location_map:
            return None
        for path in location_map[sys.platform]:
            path = Path(path).expanduser()
            if path.is_file():
                return path
        return None
